#include "common.h"
#include "io.h"
#include "pwm.h"
#include "rcservo.h"

#if defined(BC30) || defined(DJGPP)
    #include <dos.h>
#endif

RCSERVO_PARAMETER_t RCSERVO_servoParameters[32] = 
	{
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL},
		{0xffffffffL}, {0xffffffffL}, {0xffffffffL}, {0xffffffffL}
	};

unsigned long RCSERVO_playInterval = 50L;				//50ms (in ms; 20fps)


/**********************  Internal Variables & Functions  ************************/
static unsigned long RCSERVO_usedChannels;
static unsigned long RCSERVO_validChannels;  //used in Play Mode

static int RCSERVO_operationMode;
           #define RCSERVO_CAPTUREMODE	(0)
           #define RCSERVO_PLAYMODE		(1)
           #define RCSERVO_PWMMODE		(2)

/************** DX GPIO Address **********************
 * access ports of port 00--04: 0x78--0x7c
 * direction setting ports of port 00--04: 0x98--0x9c
 *     bit i = 1 => i-th pin out
 *     bit i = 0 => i-th pin in
 *****************************************************/
static unsigned GPIO_port[4]    = {0x78, 0x79, 0x7a, 0x7c};
static unsigned GPIO_dirPort[4] = {0x98, 0x99, 0x9a, 0x9c};


static void stop_PWMMode_pulses(void) {
	int i;

	for (i=0; i<32; i++)
	{
		if ((RCSERVO_usedChannels & (1L<<i)) == 0L) continue;

		if (pwm_ReadPulseCount(i) == 0L) continue; else pwm_SetPulseCount(i, 1L);
		while (pwm_ReadPulseCount(i) != 0L);
	}
}

static void stop_PlayMode_pulses(void) {
	int i;

	pwm_ClearMultiFLAG(RCSERVO_validChannels);
	for (i=0; i<32; i++)
	{
		if ((RCSERVO_validChannels & (1L<<i)) == 0L) continue;
		pwm_SetPulseCount(i, 1L);
		pwm_SetCountingMode(i, PWM_COUNT_MODE);
	}
	while (pwm_ReadMultiFLAG(RCSERVO_validChannels) != RCSERVO_validChannels);
}

static void set_PlayMode_pulses(unsigned long* width) {
	int i;

	if (RCSERVO_operationMode == RCSERVO_PLAYMODE) stop_PlayMode_pulses();

	for (i=0; i<32; i++)
	{
		if ((RCSERVO_validChannels & (1L<<i)) == 0L) continue;

		if ((width[i] < RCSERVO_servoParameters[i].minDutyWidth) || (width[i] > RCSERVO_servoParameters[i].maxDutyWidth))
			errmsg("WARNING: %s receives width %lu out of range for channel %d!\n", __FUNCTION__, width[i], i);
		else
		{
			pwm_SetPulse(i, RCSERVO_servoParameters[i].periodWidth, width[i]);
			pwm_SetCountingMode(i, PWM_CONTINUE_MODE);
		}
	} //end for i...
}
/*-------------------------  end of Internal Setting  --------------------------*/


static unsigned char RCSERVO_oldGPIODIR[4];
static unsigned char RCSERVO_oldGPIO[4];
static bool          RCSERVO_inUse = false;
bool rcservo_Initialize(unsigned long usedchannels) {
/*****************************  Assumption  *********************************
 * The function assumes that all PWM pins are disabled before being called.
 * Without this, the funciton must be written more carefully for avoiding
 * RCSERVO servos' wrong actions due to indefinite false pulses that may occur
 * when GPIO/PWM pins with indefinite states are switched to PWM outputs.
 ****************************************************************************/
	int i;
	unsigned long invalidDutyWidth = 0L;

	if (RCSERVO_inUse == true)
	{
        err_SetMsg(ERROR_RCSERVO_INUSE, "RCSERVO lib was already opened");
		return false;
	}

    //NOTE: base address should be selected carefully to avoid I/O-space conflicts!
	if (!pwm_Initialize(0xfe00, PWMCLOCK_50MHZ, PWMIRQ_DISABLE)) return false;

	RCSERVO_usedChannels = usedchannels;
	for (i=0; i<32; i++)
	{
		if (RCSERVO_servoParameters[i].periodWidth == 0xffffffffL)
			rcservo_SetServo(i, RCSERVO_DEFAULTSERVO);

		if (RCSERVO_servoParameters[i].invalidDutyWidth > invalidDutyWidth)
			invalidDutyWidth = RCSERVO_servoParameters[i].invalidDutyWidth;
	}

	//experimental trick to avoid RC servos' wrong actions
	delay_ms(1L + invalidDutyWidth/1000L);
	pwm_DisableMultiPWM(RCSERVO_usedChannels);
	pwm_EnableMultiPin(RCSERVO_usedChannels);
	for (i=0; i<32; i++)
	{
		if ((RCSERVO_usedChannels & (1L<<i)) == 0) continue;
		pwm_SetCountingMode(i, PWM_COUNT_MODE);
		pwm_SetWaveform(i, PWM_WAVEFORM_NORMAL);
		pwm_SetPulse(i, RCSERVO_servoParameters[i].periodWidth, RCSERVO_servoParameters[i].invalidDutyWidth);
		pwm_SetPulseCount(i, 0L);
	}
	delay_ms(1L + invalidDutyWidth/1000L);
	pwm_EnableMultiPWM(RCSERVO_usedChannels);

	pwm_ClearMultiFLAG(RCSERVO_usedChannels);
	for (i=0; i<32; i++)  //initialize RC servos with POWEROFF state
	{
		if ((RCSERVO_usedChannels & (1L<<i)) == 0) continue;
		pwm_SetPulse(i, RCSERVO_servoParameters[i].periodWidth, RCSERVO_servoParameters[i].cmdDuty[RCSERVO_READPOS_POWEROFF]);
		pwm_SetPulseCount(i, 1L);
	}
	while (pwm_ReadMultiFLAG(RCSERVO_usedChannels) != RCSERVO_usedChannels);


	//set those pins of GPIO ports 0,1,2,4, which correspond to the used channels, as input pins for position reading
	for (i=0; i<4; i++)
	{
		RCSERVO_oldGPIO[i] = io_inpb(GPIO_port[i]) & (unsigned char)usedchannels;    //backup GPIO output status of used channels

		RCSERVO_oldGPIODIR[i] = io_inpb(GPIO_dirPort[i]);
		io_outpb(GPIO_dirPort[i], RCSERVO_oldGPIODIR[i] & ~((unsigned char)usedchannels));
		RCSERVO_oldGPIODIR[i] = RCSERVO_oldGPIODIR[i] & (unsigned char)usedchannels; //backup GPIO direction status of used channels

		usedchannels = usedchannels>>8;
	}

	RCSERVO_operationMode = RCSERVO_CAPTUREMODE;
	RCSERVO_inUse = true;
	return true;
}

void rcservo_Close(void) {
	int i;
    unsigned long usedchannels;

	if (RCSERVO_inUse == false) return;

	//experimental trick to avoid RC servos' wrong actions
	rcservo_EnterCaptureMode();
	pwm_ClearMultiFLAG(RCSERVO_usedChannels);
	for (i=0; i<32; i++)  //set RC servos with POWEROFF state
	{
		if ((RCSERVO_usedChannels & (1L<<i)) == 0) continue;
		pwm_SetPulse(i, RCSERVO_servoParameters[i].periodWidth, RCSERVO_servoParameters[i].cmdDuty[RCSERVO_READPOS_POWEROFF]);
		pwm_SetPulseCount(i, 1L);
	}
	while (pwm_ReadMultiFLAG(RCSERVO_usedChannels) != RCSERVO_usedChannels);

	//restore GPIO output and direction setting
    usedchannels = ~RCSERVO_usedChannels;
	for (i=0; i<4; i++)
	{
		io_outpb(GPIO_port[i], (io_inpb(GPIO_port[i]) & (unsigned char)usedchannels) | RCSERVO_oldGPIO[i]);
		io_outpb(GPIO_dirPort[i], (io_inpb(GPIO_dirPort[i]) & (unsigned char)usedchannels) | RCSERVO_oldGPIODIR[i]);
		usedchannels = usedchannels>>8;
	}

	pwm_DisableMultiPin(RCSERVO_usedChannels);  //lazy code:p
	pwm_Close();
	RCSERVO_inUse = false;
}


/*****************  Primitive Position Reading Functions  *****************/
static unsigned long rcservo_ReadPosition_normal(int channel, int cmd) {
	unsigned long width;
	RCSERVO_PARAMETER_t* servo = &RCSERVO_servoParameters[channel];
	unsigned long        captureTimeOut = servo->captureMaxWidth +
							(servo->captureInitDelay + servo->captureFinishDelay) * 10L / servo->captureResolution;

	unsigned	  port = GPIO_port[channel/8];
	unsigned char pin  = 1 << (channel % 8);

    #if defined(BC30) || defined(DJGPP)
        disable();
    #endif

	//^^^^^^^^^ send command to RC servo ^^^^^^^^^^^^^^^^^^^^^^^^^^^
	pwm_SetPulse(channel, servo->cmdDuty[cmd]+servo->captureInitDelay, servo->cmdDuty[cmd]);
	pwm_SetPulseCount(channel, 1L);
    while (pwm_ReadPulseCount(channel) != 0);


	//^^^^^^^^^ read feedback of RC servo ^^^^^^^^^^^^^^^^^^^^^^^^^^
    //set the PWM channel as gpio input
	pwm_DisablePin(channel);

	//detect the raising edge
	pwm_SetPulse_10MHZ(channel, servo->captureResolution, 1L);
	pwm_SetPulseCount(channel, captureTimeOut);
    while ((io_inpb(port) & pin) == 0)
		if ((width = pwm_ReadPulseCount(channel)) == 0L)
		{   //timeout
			width = 0xffffffffL;
			goto CAPTURE_END;
		}

	//capture the falling edge
	pwm_DisablePWM(channel);
	pwm_SetPulseCount(channel, servo->captureMaxWidth);
	pwm_EnablePWM(channel);
	while ((io_inpb(port) & pin) != 0)
		if ((width = pwm_ReadPulseCount(channel)) == 0L)
		{   //timeout
			width = 0xffffffffL;
			goto CAPTURE_END;
		}

	width = (servo->captureMaxWidth - width + servo->ROBOARD_widthOffest); //in RCSERVO_captureResolution
    while ((io_inpb(port) & pin) == 0);  //should check timeout here; but I'm lazy:p

CAPTURE_END:
	//trick to avoid RC servo's wrong action in multiprogramming OS
	pwm_DisablePWM(channel);
	pwm_SetPulse(channel, servo->invalidDutyWidth + servo->minLowPhaseWidth, servo->invalidDutyWidth);
	pwm_SetPulseCount(channel, 1L);
	pwm_EnablePWM(channel);

	//set the PWM channel as PWM output
	pwm_EnablePin(channel);
    while (pwm_ReadPulseCount(channel) != 0L);

    #if defined(BC30) || defined(DJGPP)
        enable();
    #endif

	return width;
} //end rcservo_ReadPosition_normal(...)


//faster but possibly causing RC servo's wrong action in multiprogramming OS
static unsigned long rcservo_ReadPosition_fast(int channel, int cmd) {
	unsigned long width;
	RCSERVO_PARAMETER_t* servo = &RCSERVO_servoParameters[channel];
	unsigned long        captureTimeOut = servo->captureMaxWidth +
							(servo->captureInitDelay + servo->captureFinishDelay) * 10L / servo->captureResolution;

	unsigned	  dirport = GPIO_dirPort[channel/8];
	unsigned	  port    = GPIO_port[channel/8];
	unsigned char pin     = 1 << (channel % 8);

    #if defined(BC30) || defined(DJGPP)
        disable();
    #endif

	//^^^^^^^^^ send command to RC servo ^^^^^^^^^^^^^^^^^^^^^^
	pwm_SetPulse(channel, servo->cmdDuty[cmd]+servo->captureInitDelay, servo->cmdDuty[cmd]);
	pwm_SetPulseCount(channel, 1L);
    while (pwm_ReadPulseCount(channel) != 0);


	//^^^^^^^^^ read feedback of RC servo ^^^^^^^^^^^^^^^^^^^^^

	//set the PWM channel as gpio input
	pwm_DisablePin(channel); 

	//detect raising edge
	pwm_SetPulse_10MHZ(channel, servo->captureResolution, 1L);
	pwm_SetPulseCount(channel, captureTimeOut);
    while((io_inpb(port) & pin) == 0)
		if ((width = pwm_ReadPulseCount(channel)) == 0L)
		{   //timeout
			width = 0xffffffffL;
			goto CAPTURE_END;
		}

	//capture falling edge
	pwm_DisablePWM(channel);
	pwm_SetPulseCount(channel, servo->captureMaxWidth);
	pwm_EnablePWM(channel);
	while((io_inpb(port) & pin) != 0)
		if ((width = pwm_ReadPulseCount(channel)) == 0L)
		{   //timeout
			width = 0xffffffffL;
			goto CAPTURE_END;
		}

	width = (servo->captureMaxWidth - width + servo->ROBOARD_widthOffest); //in RCSERVO_captureResolution

CAPTURE_END:
	//wait RC servo to switch from "feedback" state to "PWM" state
	io_outpb(port, io_inpb(port) & (~pin));
	io_outpb(dirport, io_inpb(dirport) | pin);
	pwm_DisablePWM(channel);
	pwm_SetPulse(channel, servo->captureFinishDelay, 1L);
	pwm_SetPulseCount(channel, 1L);
	pwm_EnablePWM(channel);
    while (pwm_ReadPulseCount(channel) != 0L);

    //set the PWM channel as PWM output
	pwm_EnablePin(channel);
	io_outpb(dirport, io_inpb(dirport) & (~pin));

    #if defined(BC30) || defined(DJGPP)
        enable();
    #endif

	return width;
} //end rcservo_ReadPosition_fast(...)


unsigned long rcservo_ReadPosition(int channel, int cmd) {
	unsigned long width;

	if ((RCSERVO_usedChannels & (1L<<channel)) == 0L) return 0xffffffffL;

	if ((cmd < 0) || (cmd > 7))
	{
		errmsg("WARNING: %s() receives wrong cmd argument!\n", __FUNCTION__);
		return 0xffffffffL;
	}

	if (RCSERVO_operationMode != RCSERVO_CAPTUREMODE)
	{
		errmsg("WARNING: %s() can't work in PLAY or PWM mode!\n", __FUNCTION__);
		return 0xffffffffL;
	}

	if (RCSERVO_servoParameters[channel].feedbackReadMode == RCSERVO_READPOS_NORMALMODE)
		width = rcservo_ReadPosition_normal(channel, cmd);
	else
		width = rcservo_ReadPosition_fast(channel, cmd);

	if (width != 0xffffffffL)
	{
		width = (width < RCSERVO_servoParameters[channel].minDutyWidth)? RCSERVO_servoParameters[channel].minDutyWidth : width;
		width = (width > RCSERVO_servoParameters[channel].maxDutyWidth)? RCSERVO_servoParameters[channel].maxDutyWidth : width;
	}

	return width;
}

void rcservo_EnterCaptureMode(void) {
	if (RCSERVO_operationMode == RCSERVO_CAPTUREMODE) return;

	if (RCSERVO_operationMode == RCSERVO_PLAYMODE)
		stop_PlayMode_pulses();
	else
	if (RCSERVO_operationMode == RCSERVO_PWMMODE)
		stop_PWMMode_pulses();

	RCSERVO_operationMode = RCSERVO_CAPTUREMODE;
}
/*----------------  end Primitive Position Reading Functions  ---------------*/


/********************  Ad Hoc Position Reading Functions  ********************/
static void InsertionSort(unsigned long* A, int size) {
	unsigned long tmp;
	int i, j;

	for (i=1; i<size; i++)
	{
		tmp = A[i];
		for (j=i-1; j>=0; j--)
			if (tmp<A[j]) A[j+1]=A[j]; else break;
		A[j+1]=tmp;
	}
}

//rcservo_ReadPosition denoise version
unsigned long rcservo_ReadPositionDN(int channel, int cmd) {
	unsigned long width[100];
	int i, j=0;

	for (i=0; i<RCSERVO_servoParameters[channel].feedbackFilterWidth; i++)
	{
		for (; j<RCSERVO_servoParameters[channel].feedbackMaxNumFail; j++)
			if ((width[i] = rcservo_ReadPosition(channel, cmd)) != 0xffffffffL) break;
		if (j == RCSERVO_servoParameters[channel].feedbackMaxNumFail) return 0xffffffffL; //position read fail!
	}

	InsertionSort(width, RCSERVO_servoParameters[channel].feedbackFilterWidth);
	return width[RCSERVO_servoParameters[channel].feedbackFilterWidth/2];
}

//rcservo_ReadPosition robust version
unsigned long rcservo_ReadPositionRB(int channel, int cmd) {
    unsigned long width;
	int i;

	for (i=0; i<RCSERVO_servoParameters[channel].feedbackMaxNumFail; i++)
		if ((width = rcservo_ReadPosition(channel, cmd)) != 0xffffffffL) return width;
	return 0xffffffffL;
}


void rcservo_ReadPositions(unsigned long channels, int cmd, unsigned long* width) {
	unsigned long readchannels;
	int i;

	readchannels = RCSERVO_usedChannels & channels;
	for (i=0; i<32; i++)
		width[i] = ((readchannels & (1L<<i)) == 0L)? 0L : rcservo_ReadPositionRB(i, cmd);
}

void rcservo_ReadPositionsDN(unsigned long channels, int cmd, unsigned long* width) {
	unsigned long readchannels;
	int i;

	readchannels = RCSERVO_usedChannels & channels;
	for (i=0; i<32; i++)
		width[i] = ((readchannels & (1L<<i)) == 0L)? 0L : rcservo_ReadPositionDN(i, cmd);
}

void rcservo_PipelinedReadPositions(unsigned long channels, int cmd, unsigned long* width) {
    //TO DO (for safe but slow version)
}

void rcservo_ParallelReadPositions(unsigned long channels, int cmd, unsigned long* width) {
    //TO DO
}


static void rcservo_ReadPositions_PlayMode(unsigned long* width) { //used by rcservo_EnterPlayMode()
	int i;

	RCSERVO_validChannels = RCSERVO_usedChannels;
	for (i=0; i<32; i++)
	{
		width[i] = ((RCSERVO_usedChannels & (1L<<i)) == 0L)? 0L : rcservo_ReadPositionDN(i, RCSERVO_servoParameters[i].cmdInPlayMode);
		if (width[i] == 0xffffffffL)
		{
			errmsg("WARNING: %s() fails to read channel %d!\n", __FUNCTION__, i);
			width[i] = 0L;
			RCSERVO_validChannels = RCSERVO_validChannels & (~(1L<<i));
		}
	} //end for i ...
}
/*---------------  end of Ad Hoc Position Reading Functions  ----------------*/


/**********************  Action Playing Functions  ***************************/
static int RCSERVO_playState = RCSERVO_PLAYEND;
static unsigned long RCSERVO_curaction_starttime;
static unsigned long RCSERVO_curaction_nexttime;
static unsigned long RCSERVO_curaction_endtime;
static unsigned long RCSERVO_curaction[32];
static unsigned long RCSERVO_lastaction[32];

static unsigned long interpolate(unsigned long start_value, unsigned long end_value, unsigned long total, unsigned long t) {
	if (total == 0L) return end_value;

	return ((total-t)*start_value + t*end_value)/total;
}


void rcservo_EnterPlayMode(void) {
	int i;

	rcservo_EnterCaptureMode();
	rcservo_ReadPositions_PlayMode(RCSERVO_curaction);
	for (i=0; i<32; i++) RCSERVO_lastaction[i] = RCSERVO_curaction[i];

	set_PlayMode_pulses(RCSERVO_curaction);
	RCSERVO_operationMode = RCSERVO_PLAYMODE;
}

void rcservo_SetAction(unsigned long* width, unsigned long playtime) {
	int i;

	for (i=0; i<32; i++)
	{
		RCSERVO_lastaction[i] = RCSERVO_curaction[i];
		RCSERVO_curaction[i]  = (width[i] == 0L)? RCSERVO_curaction[i] : width[i];	    
	}

	RCSERVO_curaction_starttime = timer_nowtime();
	RCSERVO_curaction_nexttime  = RCSERVO_curaction_starttime + RCSERVO_playInterval;
	RCSERVO_curaction_endtime   = RCSERVO_curaction_starttime + playtime;

	RCSERVO_playState = RCSERVO_PLAYING;
}

int rcservo_PlayAction(void) {
	unsigned long nowtime, t, tt;
	unsigned long midaction[32];
	int i;

	if (RCSERVO_playState != RCSERVO_PLAYING) return RCSERVO_playState;

	nowtime = timer_nowtime();
	if (nowtime >= RCSERVO_curaction_endtime)
	{
		set_PlayMode_pulses(RCSERVO_curaction);
		RCSERVO_playState = RCSERVO_PLAYEND;
	}
	else
    if (nowtime >= RCSERVO_curaction_nexttime)
	{
		t  = nowtime - RCSERVO_curaction_starttime;
		tt = RCSERVO_curaction_endtime - RCSERVO_curaction_starttime;

		for (i=0; i<32; i++)
		{
			if ((RCSERVO_validChannels & (1L<<i)) == 0L) continue;
		    midaction[i] = interpolate(RCSERVO_lastaction[i], RCSERVO_curaction[i], tt, t);
		}
		set_PlayMode_pulses(midaction);

		while (RCSERVO_curaction_nexttime <= nowtime)
			RCSERVO_curaction_nexttime = RCSERVO_curaction_nexttime + RCSERVO_playInterval;
	}

	return RCSERVO_playState;
}

void rcservo_PauseAction(void) {
	unsigned long nowtime, t, tt;
	int i;

	if ((RCSERVO_playState == RCSERVO_PAUSED) || (RCSERVO_playState == RCSERVO_PLAYEND)) return;

	RCSERVO_playState = RCSERVO_PAUSED;
	if ((nowtime = timer_nowtime()) >= RCSERVO_curaction_endtime)
    {
        RCSERVO_curaction_starttime = 0xffffffffL;
        return;
    }

	t  = nowtime - RCSERVO_curaction_starttime;
	tt = RCSERVO_curaction_endtime - RCSERVO_curaction_starttime;

	for (i=0; i<32; i++)
	{
		if ((RCSERVO_validChannels & (1L<<i)) == 0L) continue;
	    RCSERVO_lastaction[i] = interpolate(RCSERVO_lastaction[i], RCSERVO_curaction[i], tt, t);
	}
	RCSERVO_curaction_starttime = RCSERVO_curaction_endtime - nowtime;
}

void rcservo_ReleaseAction(void) {
	unsigned long nowtime;

	if (RCSERVO_playState != RCSERVO_PAUSED) return;

	RCSERVO_playState = RCSERVO_PLAYING;
	if (RCSERVO_curaction_starttime == 0xffffffffL) return;

	nowtime = timer_nowtime();
	RCSERVO_curaction_endtime   = RCSERVO_curaction_starttime + nowtime;
	RCSERVO_curaction_starttime = nowtime;
}
/*---------------------  end of Action Playing Functions  --------------------*/


/***************************  PWM Mode Functions  *****************************/
void rcservo_EnterPWMMode(void) {
	rcservo_EnterCaptureMode();
	RCSERVO_operationMode = RCSERVO_PWMMODE;
}

void rcservo_SendPWMPulses(int channel, unsigned long period, unsigned long duty, unsigned long count) {
	if ((RCSERVO_usedChannels & (1L<<channel)) == 0L) return;

	if ((period == 0L) || (duty == 0L))
	{
		errmsg("WARNING: %s() receives a zero period or duty width!\n", __FUNCTION__);
		return;
	}

	if (count > 0x0fffffffL)
	{
		errmsg("WARNING: %s() receives a too large pulse count!\n", __FUNCTION__);
		return;
	}

	pwm_SetPulse(channel, period, duty);
	pwm_SetPulseCount(channel, count);
}

bool rcservo_IsPWMCompleted(int channel) {
	if ((RCSERVO_usedChannels & (1L<<channel)) == 0L) return true;
	if (pwm_ReadPulseCount(channel) == 0L) return true; else return false;
}
/*------------------------  end of PWM Mode Functions  -----------------------*/


void rcservo_SetServo(int channel, int servono) {
	if (RCSERVO_inUse == true)
	{
		errmsg("WARNING: %s() can't set servo parameters when RCSERVO lib is in use!\n", __FUNCTION__);
		return;
	}
	if ((channel < 0) || (channel > 31))
	{
		errmsg("WARNING: %s() receives a wrong channel!\n", __FUNCTION__);
		return;
	}

	switch (servono)
	{
		case RCSERVO_DEFAULTSERVO:
			RCSERVO_servoParameters[channel].periodWidth		= 10000L;
			RCSERVO_servoParameters[channel].minDutyWidth		= 350L;
			RCSERVO_servoParameters[channel].maxDutyWidth		= 3500L;
			RCSERVO_servoParameters[channel].invalidDutyWidth	= 5000L;
			RCSERVO_servoParameters[channel].minLowPhaseWidth   = 500L;

			RCSERVO_servoParameters[channel].captureResolution	 = 10L;
			RCSERVO_servoParameters[channel].captureInitDelay    = 35L;
			RCSERVO_servoParameters[channel].captureFinishDelay  = 300L;
			RCSERVO_servoParameters[channel].captureMaxWidth     = 3500L;
			RCSERVO_servoParameters[channel].ROBOARD_widthOffest = 29L;

			RCSERVO_servoParameters[channel].cmdDuty[0] = 50L;
			RCSERVO_servoParameters[channel].cmdDuty[1] = 100L;
			RCSERVO_servoParameters[channel].cmdDuty[2] = 150L;
			RCSERVO_servoParameters[channel].cmdDuty[3] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[4] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[5] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[6] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[7] = 200L;
			RCSERVO_servoParameters[channel].cmdInPlayMode = RCSERVO_READPOS_POWEROFF;

			RCSERVO_servoParameters[channel].feedbackReadMode    = RCSERVO_READPOS_NORMALMODE;
			RCSERVO_servoParameters[channel].feedbackFilterWidth = 3;
			RCSERVO_servoParameters[channel].feedbackMaxNumFail  = 3;

			break;
		case RCSERVO_KONDO_GENERIC:
		case RCSERVO_KONDO_KRS786:
		case RCSERVO_KONDO_KRS788:
		case RCSERVO_KONDO_KRS78X:
			RCSERVO_servoParameters[channel].periodWidth		= 10000L;
			RCSERVO_servoParameters[channel].minDutyWidth		= 700L;
			RCSERVO_servoParameters[channel].maxDutyWidth		= 2300L;
			RCSERVO_servoParameters[channel].invalidDutyWidth	= 3000L;
			RCSERVO_servoParameters[channel].minLowPhaseWidth   = 200L;

			RCSERVO_servoParameters[channel].captureResolution	 = 10L;
			RCSERVO_servoParameters[channel].captureInitDelay    = 35L;
			RCSERVO_servoParameters[channel].captureFinishDelay  = 150L;
			RCSERVO_servoParameters[channel].captureMaxWidth     = 2800L;
			RCSERVO_servoParameters[channel].ROBOARD_widthOffest = 29L;

			RCSERVO_servoParameters[channel].cmdDuty[0] = 50L;
			RCSERVO_servoParameters[channel].cmdDuty[1] = 100L;
			RCSERVO_servoParameters[channel].cmdDuty[2] = 150L;
			RCSERVO_servoParameters[channel].cmdDuty[3] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[4] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[5] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[6] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[7] = 200L;
			RCSERVO_servoParameters[channel].cmdInPlayMode = RCSERVO_READPOS_USEPARAMETER1;

			RCSERVO_servoParameters[channel].feedbackReadMode    = RCSERVO_READPOS_NORMALMODE;
			RCSERVO_servoParameters[channel].feedbackFilterWidth = 3;
			RCSERVO_servoParameters[channel].feedbackMaxNumFail  = 3;

			break;
		case RCSERVO_HITEC_GENERIC:
		case RCSERVO_HITEC_HSR8498:
			RCSERVO_servoParameters[channel].periodWidth		= 10000L;
			RCSERVO_servoParameters[channel].minDutyWidth		= 550L;
			RCSERVO_servoParameters[channel].maxDutyWidth		= 2450L;
			RCSERVO_servoParameters[channel].invalidDutyWidth	= 3500L;
			RCSERVO_servoParameters[channel].minLowPhaseWidth   = 200L;

			RCSERVO_servoParameters[channel].captureResolution	 = 10L;
			RCSERVO_servoParameters[channel].captureInitDelay    = 25L;
			RCSERVO_servoParameters[channel].captureFinishDelay  = 250L;
			RCSERVO_servoParameters[channel].captureMaxWidth     = 2800L;
			RCSERVO_servoParameters[channel].ROBOARD_widthOffest = 29L;

			RCSERVO_servoParameters[channel].cmdDuty[0] = 50L;
			RCSERVO_servoParameters[channel].cmdDuty[1] = 100L;
			RCSERVO_servoParameters[channel].cmdDuty[2] = 150L;
			RCSERVO_servoParameters[channel].cmdDuty[3] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[4] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[5] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[6] = 200L;
			RCSERVO_servoParameters[channel].cmdDuty[7] = 200L;
			RCSERVO_servoParameters[channel].cmdInPlayMode = RCSERVO_READPOS_POWEROFF;

			RCSERVO_servoParameters[channel].feedbackReadMode    = RCSERVO_READPOS_NORMALMODE;
			RCSERVO_servoParameters[channel].feedbackFilterWidth = 3;
			RCSERVO_servoParameters[channel].feedbackMaxNumFail  = 3;

			break;
		default:
			errmsg("WARNING: %s() can't recognize the given servo no.!\n", __FUNCTION__);
			break;
	}//end switch (servono)
}


void rcservo_Outp(int channel, int value) {
	unsigned	  port = GPIO_port[channel/8], dirport = GPIO_dirPort[channel/8];
	unsigned char pin  = 1 << (channel % 8);

	if ((value > 1) || ((RCSERVO_usedChannels & (1L<<channel)) == 0L)) return;
	io_outpb(dirport, io_inpb(dirport) | (1 << pin));
	io_outpb(port, (io_inpb(port) & (~(1 << pin))) | (value << pin));
}


int rcservo_Inp(int channel) {
	unsigned	  port = GPIO_port[channel/8], dirport = GPIO_dirPort[channel/8];
	unsigned char pin  = 1 << (channel % 8);

	if ((RCSERVO_usedChannels & (1L<<channel)) == 0L) return 0xffff;
	io_outpb(dirport, io_inpb(dirport) & (~(1 << pin)));
    if ((io_inpb(port) & (1 << pin)) != 0) return 1; else return 0;
}

